import json
from typing import Dict, Any, List, Optional


class MockResponse:
    def __init__(
        self, *, response: Dict[str, Any], index: int, file_path: Optional[str]
    ) -> None:
        assert index is not None
        assert response is not None
        assert isinstance(response, dict)
        raw_body = response.get("body")
        if isinstance(raw_body, dict):
            raw_body = json.dumps(raw_body)
        self.raw_body: Optional[str] = raw_body
        self.json_body: Dict[str, Any] | List[Dict[str, Any]] | None
        if not self.raw_body:
            self.json_body = None
        try:
            self.json_body = json.loads(self.raw_body) if self.raw_body else None
        except json.decoder.JSONDecodeError:
            self.json_body = None

        self.status_code: int | None = response.get("statusCode")
        self.index: int = index
        self.file_path: Optional[str] = file_path



import json
from typing import Dict, Any, Optional, List, Union, cast
from urllib.parse import parse_qs

from mockserver_client.mock_request_logger import MockRequestLogger


class MockRequest:
    def __init__(
        self, request: Dict[str, Any], index: int, file_path: Optional[str]
    ) -> None:
        """
        Class for mock requests

        :param request:
        """
        assert index is not None
        self.index: int = index
        assert request is not None
        assert isinstance(request, dict)
        self.request: Dict[str, Any] = request

        self.file_path: Optional[str] = file_path

        self.method: Optional[str] = self.request.get("method")
        self.path: Optional[str] = self.request.get("path")
        self.querystring_params: Dict[str, Any] | List[
            Dict[str, Any]
        ] | None = self.request.get("queryStringParameters")
        assert (
            not self.querystring_params
            or isinstance(self.querystring_params, dict)
            or isinstance(self.querystring_params, list)
        ), type(self.querystring_params)

        self.headers: Optional[Dict[str, Any]] = self.request.get("headers")

        raw_body: str | bytes | Dict[str, Any] | List[Dict[str, Any]] = cast(
            str | bytes | Dict[str, Any] | List[Dict[str, Any]],
            self.request.get("body"),
        )

        self.body_list: Optional[List[Dict[str, Any]]] = MockRequest.parse_body(
            body=raw_body, headers=self.headers
        )

        assert self.body_list is None or isinstance(
            self.body_list, list
        ), f"{type(self.body_list)}: {json.dumps(self.body_list)}"

        raw_json_content: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = (
            self.body_list[0].get("json")
            if self.body_list is not None
            and len(self.body_list) > 0
            and "json" in self.body_list[0]
            else self.body_list
            if self.body_list is not None and len(self.body_list) > 0
            else None
        )

        self.json_list: Optional[List[Dict[str, Any]]] = (
            MockRequest.parse_body(body=raw_json_content, headers=self.headers)
            if raw_json_content
            else None
        )

        assert self.json_list is None or isinstance(
            self.json_list, list
        ), f"{type(self.json_list)}: {json.dumps(self.json_list)}"

    @staticmethod
    def parse_body(
        *,
        body: Union[str, bytes, Dict[str, Any], List[Dict[str, Any]]],
        headers: Optional[Dict[str, Any]],
    ) -> Optional[List[Dict[str, Any]]]:
        # body can be either:
        # 0. None
        # 1. bytes (UTF-8 encoded)
        # 2. str (form encoded)
        # 3. str (json)
        # 3. dict
        # 4. list of string
        # 5. list of dict

        if body is None:
            return None

        if isinstance(body, bytes):
            return MockRequest.parse_body(body=body.decode("utf-8"), headers=headers)

        if isinstance(body, str):
            return MockRequest.parse_body(body=json.loads(body), headers=headers)

        if isinstance(body, dict):
            if MockRequest.is_request_content_type_form_urlencoded(body, headers):
                return [MockRequest.convert_query_parameters_to_dict(body["string"])]
            else:
                return [body]

        if isinstance(body, list):
            my_list: List[Optional[List[Dict[str, Any]]]] = [
                MockRequest.parse_body(body=c, headers=headers)
                for c in body
                if c is not None
            ]
            return [
                item
                for sublist in my_list
                if sublist is not None
                for item in sublist
                if item is not None
            ]

        assert False, f"body is in unexpected type: {type(body)}"

    @staticmethod
    def is_request_content_type_form_urlencoded(
        body: Dict[str, Any],
        headers: Optional[Union[List[Dict[str, Any]], Dict[str, Any]]],
    ) -> bool:
        """
        check the body and headers to see if this is a form urlencoded request, it is
        the body contains "string" and the headers has Content-Type of "application/x-www-form-urlencoded"
        """
        if body and "string" in body and headers:
            # sometimes headers is a list[dict] and sometimes a dict
            headers_dict = headers[0] if isinstance(headers, list) else headers
            # sometimes headers_dict will be {"name": "Content-Type", "values": ["application/x-www-form-urlencoded"]}
            # and sometimes {"Content-Type": ["application/x-www-form-urlencoded"]}
            # fmt: off
            if "application/x-www-form-urlencoded" in headers_dict.get("Content-Type", []):
                return True
            if (headers_dict.get("name") == "Content-Type" 
                    and "application/x-www-form-urlencoded" in headers_dict.get("values", [])):
                return True
            # fmt: on
        return False

    def __str__(self) -> str:
        return f"({self.index})" f" {self.path}{MockRequestLogger.convert_query_parameters_to_str(self.querystring_params)}" + (
            f": {self.json_list}" if self.json_list else ""
        ) + (
            f" from ({self.file_path})" if self.file_path else ""
        )

    @staticmethod
    def convert_query_parameters_to_dict(query: str) -> Dict[str, str]:
        params: Dict[str, List[str]] = parse_qs(query)
        return {k: v[0] for k, v in params.items()}

    def matches_without_body(self, other: "MockRequest") -> bool:
        return (
            self.method == other.method
            and self.path == other.path
            and self.querystring_params == other.querystring_params
        )



from glob import glob
from pathlib import Path
from typing import List, Any, Dict

import pytest
import requests
import json

from requests import Response

from mockserver_client.exceptions.mock_server_expectation_not_found_exception import (
    MockServerExpectationNotFoundException,
)
from mockserver_client.exceptions.mock_server_json_content_mismatch_exception import (
    MockServerJsonContentMismatchException,
)
from mockserver_client.mockserver_client import MockServerFriendlyClient
from mockserver_client.mockserver_verify_exception import MockServerVerifyException


def test_mock_server_from_file_multiple_calls_mismatch() -> None:
    expectations_dir: Path = Path(__file__).parent.joinpath("./expectations")
    requests_dir: Path = Path(__file__).parent.joinpath("./requests")

    test_name = "test_mock_server"

    mock_server_url = "http://mock-server:1080"
    mock_client: MockServerFriendlyClient = MockServerFriendlyClient(
        base_url=mock_server_url
    )

    mock_client.clear(f"/{test_name}/*")
    mock_client.reset()

    mock_client.expect_files_as_json_requests(
        expectations_dir, path=f"/{test_name}/foo/1/merge", json_response_body={}
    )
    mock_client.expect_default()

    http = requests.Session()

    file_path: str
    files: List[str] = sorted(
        glob(str(requests_dir.joinpath("**/*.json")), recursive=True)
    )
    for file_path in files:
        with open(file_path, "r") as file:
            content: Dict[str, Any] = json.loads(file.read())
            response: Response = http.post(
                mock_server_url + "/" + test_name + "/foo/1/merge",
                json=[content],
            )
            assert response.ok

    with pytest.raises(MockServerVerifyException):
        try:
            mock_client.verify_expectations(test_name=test_name)
        except MockServerVerifyException as e:
            # there should be two expectations.
            # One for the content not matching and one for the expectation not triggered
            assert len(e.exceptions) == 2
            json_content_mismatch_exceptions: List[
                MockServerJsonContentMismatchException
            ] = [
                e1
                for e1 in e.exceptions
                if isinstance(e1, MockServerJsonContentMismatchException)
            ]
            assert len(json_content_mismatch_exceptions) == 1
            expectation_not_found_exceptions: List[
                MockServerExpectationNotFoundException
            ] = [
                e1
                for e1 in e.exceptions
                if isinstance(e1, MockServerExpectationNotFoundException)
            ]
            assert len(expectation_not_found_exceptions) == 1
            print(str(e))
            raise e



import json
from typing import Any, Dict, List, Optional

from .mock_server_exception import MockServerException
from ..mock_request import MockRequest


class MockServerRequestNotFoundException(MockServerException):
    """
    Exception when we found no expectation for a request
    """

    def __init__(
        self,
        *,
        method: Optional[str],
        url: Optional[str],
        json_list: Optional[List[Dict[str, Any]]],
        querystring_params: Dict[str, Any] | List[Dict[str, Any]] | None = None,
        request: MockRequest,
    ) -> None:
        """
        Exception when we found no expectation for a request

        :param method: method of request
        :param url: url of request
        :param json_list: json body
        :param request: request
        """
        self.method: Optional[str] = method
        self.url: Optional[str] = url
        self.json_dict: Optional[List[Dict[str, Any]]] = json_list
        self.querystring_params: Dict[str, Any] | List[
            Dict[str, Any]
        ] | None = querystring_params
        self.request: MockRequest = request
        assert (
            not json_list or isinstance(json_list, dict) or isinstance(json_list, list)
        ), type(json_list)
        super().__init__(
            f"Request was not expected: {method} {url} {querystring_params!r} "
            + f"{json.dumps(json_list) if json_list else '(No body)'}"
        )



from glob import glob
from pathlib import Path
from typing import List, Any, Dict

import requests
import json

from requests import Response

from mockserver_client.mockserver_client import MockServerFriendlyClient
from mockserver_client.mockserver_verify_exception import MockServerVerifyException


def test_mock_server_from_file_multiple_calls() -> None:
    requests_dir: Path = Path(__file__).parent.joinpath("./request_json_calls")
    test_name = "test_mock_server"

    mock_server_url = "http://mock-server:1080"
    mock_client: MockServerFriendlyClient = MockServerFriendlyClient(
        base_url=mock_server_url
    )

    mock_client.clear(f"/{test_name}/*")
    mock_client.reset()

    mock_client.expect_files_as_json_requests(
        requests_dir, path=f"/{test_name}/foo/1/merge", json_response_body={}
    )
    mock_client.expect_default()

    http = requests.Session()

    file_path: str
    files: List[str] = sorted(
        glob(str(requests_dir.joinpath("**/*.json")), recursive=True)
    )
    for file_path in files:
        with open(file_path, "r") as file:
            content: Dict[str, Any] = json.loads(file.read())
            response: Response = http.post(
                mock_server_url + "/" + test_name + "/foo/1/merge",
                json=[content],
            )
            assert response.ok

    try:
        mock_client.verify_expectations(test_name=test_name)
    except MockServerVerifyException as e:
        print(str(e))
        raise e



from pathlib import Path

import requests

from mockserver_client.mock_requests_loader import load_mock_source_api_json_responses
from mockserver_client.mockserver_client import MockServerFriendlyClient
from mockserver_client.mockserver_verify_exception import MockServerVerifyException


def test_mock_server_from_file_content_type_form_urlencoded() -> None:
    requests_dir: Path = Path(__file__).parent.joinpath("./expectations")
    test_name = "test_mock_server"

    mock_server_url = "http://mock-server:1080"
    mock_client: MockServerFriendlyClient = MockServerFriendlyClient(
        base_url=mock_server_url
    )

    mock_client.clear(f"/{test_name}/*.*")
    mock_client.reset()

    load_mock_source_api_json_responses(
        folder=requests_dir,
        mock_client=mock_client,
        url_prefix=test_name,
    )

    http = requests.Session()
    # expectation file content_type_form_urlencoded_string_body
    # this expectation is set up in the preferred way for "Content-Type": "application/x-www-form-urlencoded"
    matched_response = http.post(
        mock_server_url + "/" + test_name,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={"member_id": "12345", "status": "active"},
    )
    assert matched_response.status_code == 200

    # expectation file content_type_form_urlencoded_not_matched
    # this is a fun one because mock-server will not match the request because the service_slug value is being
    # coerced to an integer by mock-server when doing the matching. this seems to only happen for requests
    # with "Content-Type": "application/x-www-form-urlencoded
    not_found_response = http.post(
        mock_server_url + "/" + test_name + "/api/v1.0/update_token_status",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={
            "member_id": "xy18376",
            "service_slug": "2222",
            "status": "Retrieving Data",
        },
    )
    assert (
        not_found_response.status_code == 404
    ), "mock server x-www-form-urlencoded issue is resolved!"

    # expectation file content_type_form_urlencoded_matched
    # this mocked request is the same as the previous except the service_slug value is now a string due to the
    # inclusion of an alphanumeric character. this request will match the mock-server expectation
    matched_response = http.post(
        mock_server_url + "/" + test_name + "/api/v1.0/update_token_status",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={
            "member_id": "xy18376",
            "service_slug": "2222a",
            "status": "Retrieving Data",
        },
    )

    assert matched_response.status_code == 200

    try:
        mock_client.verify_expectations(test_name=test_name)
    except MockServerVerifyException as e:
        print(str(e))
        raise e



import json
from typing import Optional, Dict, Any, List

from .mock_server_exception import MockServerException
from ..mock_request import MockRequest


class MockServerExpectationNotFoundException(MockServerException):
    """
    Exception when a request was made but an expectation was not found for it
    """

    def __init__(
        self,
        *,
        method: Optional[str],
        url: Optional[str],
        json_list: Optional[List[Dict[str, Any]]],
        querystring_params: Dict[str, Any] | List[Dict[str, Any]] | None = None,
        expectation: MockRequest,
    ) -> None:
        """
        Exception when a request was made but an expectation was not found for it


        :param url: url of expectation not found
        :param json_list: json body
        :param querystring_params: query string
        :param expectation
        """
        self.method: Optional[str] = method
        self.url: Optional[str] = url
        self.json_list: Optional[List[Dict[str, Any]]] = json_list
        self.querystring_params: Dict[str, Any] | List[
            Dict[str, Any]
        ] | None = querystring_params
        self.expectation: MockRequest = expectation
        super().__init__(
            f"Expectation not met: {method} {url} {querystring_params!r} "
            + f"{json.dumps(json_list) if json_list else '(No body)'}"
        )



from typing import Dict, Any, Optional, List


class MockRequestLogger:
    @staticmethod
    def log(
        *,
        file_path: Optional[str],
        base_url: str,
        request: Dict[str, Any],
        response: Dict[str, Any] | None = None,
    ) -> None:
        """
        Prints the mock request

        :param base_url: base url
        :param file_path: file name
        :param request: mock request
        :param response: mock response
        """
        method: str = request.get("method", "GET")
        path: Optional[str] = request.get("path")
        query_string_parameters: Optional[Dict[str, Any]] = request.get(
            "queryStringParameters"
        )
        body: Optional[Dict[str, Any]] = request.get("body")
        response_body: Optional[Dict[str, Any]] = (
            response.get("body") if response else None
        )

        print(
            f"Mocking: {method} {base_url}{path}"
            + (
                f"{MockRequestLogger.convert_query_parameters_to_str(query_string_parameters)}"
                if query_string_parameters
                else ""
            )
            + (f", from: ({file_path})" if file_path else "")
            + (f", body: {body}" if body else "")
            + (f", response: {response_body}" if response_body else "")
        )

    @staticmethod
    def convert_query_parameters_to_str(
        query_parameters: Dict[str, Any] | List[Dict[str, Any]] | None
    ) -> str:
        if query_parameters is None:
            return ""
        if isinstance(query_parameters, dict):
            return "?" + "&".join(
                [
                    f"{k}={MockRequestLogger.get_value_as_str(v)}"
                    for k, v in query_parameters.items()
                ]
            )
        assert isinstance(query_parameters, list)
        return "?" + "&".join(
            [
                f"{v.get('name')}={MockRequestLogger.get_value_as_str(v.get('values'))}"
                for v in query_parameters
            ]
        )

    @staticmethod
    def get_value_as_str(values: List[Any] | None) -> str:
        if values is not None:
            assert isinstance(values, list)
            return ",".join([str(v) for v in values])
        return ""



from typing import Dict, Any

from mockserver_client.mock_request import MockRequest
from mockserver_client.mock_response import MockResponse


class MockRequestResponse:
    def __init__(
        self,
        *,
        request: Dict[str, Any] | None,
        response: Dict[str, Any] | None,
        index: int,
        file_path: str | None = None,
    ):
        self.raw_request: Dict[str, Any] | None = request
        self.request: MockRequest | None = (
            MockRequest(request=request, index=index, file_path=file_path)
            if request
            else None
        )
        self.raw_response: Dict[str, Any] | None = response
        self.response: MockResponse | None = (
            MockResponse(response=response, index=index, file_path=file_path)
            if response
            else None
        )
        self.file_path: str | None = file_path



import json
from pathlib import Path
from typing import Any, Dict, List, Optional

from .mock_server_exception import MockServerException


class MockServerJsonContentMismatchException(MockServerException):
    """
    Exception when a request was made and an expectation with the same url was found
        but the content of the request did not match the content of the expectation
    """

    def __init__(
        self,
        actual_json: Optional[List[Dict[str, Any]]],
        expected_json: Optional[List[Dict[str, Any]]],
        differences: List[str],
        expected_file_path: Path,
    ) -> None:
        """
        Exception when a request was made and an expectation with the same url was found
            but the content of the request did not match the content of the expectation

        :param actual_json: json of actual request
        :param expected_json: json of expected request
        :param differences: differences
        :param expected_file_path:
        """
        self.actual_json: Optional[List[Dict[str, Any]]] = actual_json
        assert isinstance(actual_json, list), type(actual_json)
        self.expected_json: Optional[List[Dict[str, Any]]] = expected_json
        assert isinstance(expected_json, list), type(expected_json)
        self.differences: List[str] = differences
        assert isinstance(differences, list), type(differences)
        self.expected_file_path = expected_file_path
        assert isinstance(expected_file_path, Path), type(expected_file_path)
        error_message: str = f"Expected vs Actual: {differences} [{expected_file_path}]"
        if expected_json is None and actual_json is not None:
            error_message = f"Expected was None but Actual is {json.dumps(actual_json)}"
        elif expected_json is not None and actual_json is None:
            error_message = (
                f"Expected was {json.dumps(expected_json)} but Actual is None"
            )
        elif (
            expected_json is not None
            and actual_json is not None
            and len(self.actual_json) != len(self.expected_json)  # type: ignore
        ):
            error_message = f"Expected has {len(expected_json)} rows while actual has {len(actual_json)} rows"
        super().__init__(error_message)



import dataclasses
from typing import List

from mockserver_client.exceptions.mock_server_exception import MockServerException
from mockserver_client.mock_request import MockRequest


@dataclasses.dataclass
class MatchRequestResult:
    exceptions: List[MockServerException]
    found_expectations: List[MockRequest]



import json
from pathlib import Path
from typing import Any, Dict, List, Optional

from .exceptions.mock_server_exception import MockServerException
from .exceptions.mock_server_json_content_mismatch_exception import (
    MockServerJsonContentMismatchException,
)
from .mock_request import MockRequest


class MockServerVerifyException(Exception):
    def __init__(
        self,
        exceptions: List[MockServerException],
        found_expectations: List[MockRequest],
        files: Optional[List[str]] = None,
    ) -> None:
        """
        Exception when mock server found the mismatch

        :param exceptions:
        :param files:
        """
        super().__init__()
        self.exceptions: List[MockServerException] = exceptions
        self.found_expectations: List[MockRequest] = found_expectations
        self.files: List[str] = files or []
        self.set_files_in_exceptions()

    def set_files_in_exceptions(self) -> None:
        # iterate through the exceptions and find the corresponding files
        exception: MockServerException
        for exception in self.exceptions:
            if isinstance(exception, MockServerJsonContentMismatchException):
                actual_list_or_obj: Optional[
                    List[Dict[str, Any]]
                ] = exception.actual_json
                if actual_list_or_obj:
                    for actual in actual_list_or_obj:
                        if "resourceType" in actual and "id" in actual:
                            # noinspection PyPep8Naming
                            resourceType: str = actual["resourceType"]
                            id_: str = actual["id"]
                            # now iterate through the files to find the file for this
                            file_name: str
                            for file_name in self.files:
                                with open(file_name, "r") as file:
                                    json_data = json.loads(file.read())
                                    if isinstance(json_data, list):
                                        for json_data1 in json_data:
                                            if (
                                                json_data1["resourceType"]
                                                == resourceType
                                                and json_data1["id"] == id_
                                            ):
                                                exception.expected_file_path = Path(
                                                    file_name
                                                )
                                    else:
                                        if (
                                            json_data["resourceType"] == resourceType
                                            and json_data["id"] == id_
                                        ):
                                            exception.expected_file_path = Path(
                                                file_name
                                            )

    def __str__(self) -> str:
        return ",".join(
            [
                str(e.expected_file_path)
                for e in self.exceptions
                if isinstance(e, MockServerJsonContentMismatchException)
            ]
        )



import json
import os
from glob import glob
from pathlib import Path
from typing import Dict, Any, List, Optional

from mockserver_client.mockserver_client import (
    mock_request,
    mock_response,
    MockServerFriendlyClient,
    json_equals,
    times,
    text_equals,
)


def load_mock_fhir_requests_from_folder(
    folder: Path,
    mock_client: MockServerFriendlyClient,
    method: str = "POST",
    relative_path: Optional[str] = None,
    query_string: Optional[Dict[str, Any]] = None,
    url_prefix: Optional[str] = None,
    response_body: Optional[str] = None,
) -> List[str]:
    """
    Loads all .json files from the folder and its sub-folders

    from https://pypi.org/project/mockserver-friendly-client/

    :param folder: where to look for .json files (recursively)
    :param mock_client: client to mock server
    :param method:
    :param relative_path:
    :param query_string:
    :param url_prefix:
    :param response_body:
    """
    file_path: str
    files: List[str] = sorted(
        glob(str(folder.joinpath("**/*.json")), recursive=True), reverse=True
    )
    for file_path in files:
        # load file as json
        with open(file_path, "r") as file:
            contents = json.loads(file.read())
            if isinstance(contents, list) and not relative_path:
                for fhir_request in contents:
                    mock_single_request(
                        fhir_request=fhir_request,
                        method=method,
                        mock_client=mock_client,
                        relative_path=relative_path,
                        query_string=query_string,
                        url_prefix=url_prefix,
                        response_body=response_body,
                        file_path=file_path,
                    )
            elif contents.get("resourceType") == "Bundle" and contents.get("entry"):
                mock_bundle_request(
                    fhir_request=contents,
                    method=method,
                    mock_client=mock_client,
                    relative_path=relative_path,
                    query_string=query_string,
                    url_prefix=url_prefix,
                    response_body=response_body,
                    bundle=contents,
                    file_path=file_path,
                )
            else:
                mock_single_request(
                    fhir_request=contents,
                    method=method,
                    mock_client=mock_client,
                    relative_path=relative_path,
                    query_string=query_string,
                    url_prefix=url_prefix,
                    response_body=response_body,
                    file_path=file_path,
                )

    return files


def load_mock_fhir_requests_for_single_file(
    folder: Path,
    single_file_name: str,
    mock_client: MockServerFriendlyClient,
    method: str = "POST",
    relative_path: Optional[str] = None,
    query_string: Optional[Dict[str, Any]] = None,
    url_prefix: Optional[str] = None,
    response_body: Optional[str] = None,
) -> List[str]:
    """
    Loads a single .json file from the given folder

    from https://pypi.org/project/mockserver-friendly-client/

    :param folder: where to look for a given .json file
    :param single_file_name: a single json file name
    :param mock_client: client to mock server
    :param method:
    :param relative_path:
    :param query_string:
    :param url_prefix:
    :param response_body:
    """

    file_path: str
    files: List[str] = sorted(
        glob(str(folder.joinpath(f"**/{single_file_name}")), recursive=True)
    )
    for file_path in files:
        # load file as json
        with open(file_path, "r") as file:
            contents = json.loads(file.read())

            mock_single_request(
                fhir_request=contents,
                method=method,
                mock_client=mock_client,
                relative_path=relative_path,
                query_string=query_string,
                url_prefix=url_prefix,
                response_body=response_body,
                file_path=file_path,
            )

    return files


def mock_single_request(
    fhir_request: Dict[str, Any],
    method: str,
    mock_client: MockServerFriendlyClient,
    relative_path: Optional[str],
    query_string: Optional[Dict[str, Any]],
    url_prefix: Optional[str],
    response_body: Optional[str],
    file_path: Optional[str],
) -> None:
    # find id and resourceType
    if method == "POST":
        # id_ = fhir_request["id"]
        # noinspection PyPep8Naming
        resourceType = fhir_request["resourceType"]
        id_ = fhir_request["id"]
        path = f"{('/' + url_prefix) if url_prefix else ''}/4_0_0/{resourceType}/{id_}/$merge"
        payload: str = (
            json.dumps(
                [
                    {
                        "id": id_,
                        "updated": False,
                        "created": True,
                        "resourceType": resourceType,
                    }
                ]
            )
            if not response_body
            else response_body
        )
        mock_client.expect(
            request=mock_request(
                method="POST",
                path=path,
                body=json_equals(fhir_request),
            ),
            response=mock_response(body=payload),
            timing=times(1),
            file_path=file_path,
        )
    else:
        if not relative_path:
            id_ = fhir_request["id"]
            # noinspection PyPep8Naming
            resourceType = fhir_request["resourceType"]
            path = (
                f"{('/' + url_prefix) if url_prefix else ''}/4_0_0/{resourceType}/{id_}"
            )
            mock_client.expect(
                request=mock_request(method="GET", path=path, querystring=query_string),
                response=mock_response(body=json.dumps(fhir_request)),
                timing=times(1),
                file_path=file_path,
            )
        else:
            path = f"{('/' + url_prefix) if url_prefix else ''}/4_0_0/{relative_path}"
            mock_client.expect(
                request=mock_request(method="GET", path=path, querystring=query_string),
                response=mock_response(body=json.dumps(fhir_request)),
                timing=times(1),
                file_path=file_path,
            )


def mock_bundle_request(
    fhir_request: Dict[str, Any],
    method: str,
    mock_client: MockServerFriendlyClient,
    relative_path: Optional[str],
    query_string: Optional[Dict[str, Any]],
    url_prefix: Optional[str],
    response_body: Optional[str],
    bundle: Dict[str, Any],
    file_path: Optional[str],
) -> None:
    # find id and resourceType
    if method == "POST":
        id_ = fhir_request["id"] or "1"
        # noinspection PyPep8Naming
        resourceType = fhir_request["resourceType"]
        path = f"{('/' + url_prefix) if url_prefix else ''}/4_0_0/{resourceType}/{id_}/$merge"
        bundle_entries = bundle.get("entry")
        payload: str = (
            json.dumps(
                [
                    {
                        "id": entry.get("resource", {}).get("id", ""),
                        "updated": False,
                        "created": True,
                        "resourceType": entry.get("resource", {}).get(
                            "resourceType", ""
                        ),
                    }
                    for entry in bundle_entries
                ]
                if bundle_entries
                else []
            )
            if not response_body
            else response_body
        )
        mock_client.expect(
            request=mock_request(
                method="POST",
                path=path,
                body=json_equals(fhir_request),
            ),
            response=mock_response(body=payload),
            timing=times(1),
            file_path=file_path,
        )
    else:
        if not relative_path:
            id_ = fhir_request["id"]
            # noinspection PyPep8Naming
            resourceType = fhir_request["resourceType"]
            path = (
                f"{('/' + url_prefix) if url_prefix else ''}/4_0_0/{resourceType}/{id_}"
            )
            mock_client.expect(
                request=mock_request(method="GET", path=path, querystring=query_string),
                response=mock_response(body=json.dumps(fhir_request)),
                timing=times(1),
                file_path=file_path,
            )
        else:
            path = f"{('/' + url_prefix) if url_prefix else ''}/4_0_0/{relative_path}"
            mock_client.expect(
                request=mock_request(method="GET", path=path, querystring=query_string),
                response=mock_response(body=json.dumps(fhir_request)),
                timing=times(1),
                file_path=file_path,
            )


# noinspection PyPep8Naming
def load_mock_fhir_everything_requests_from_folder(
    folder: Path,
    mock_client: MockServerFriendlyClient,
    resourceType: str,
    url_prefix: Optional[str] = None,
) -> List[str]:
    """
    Loads all .json files from the folder and its sub-folders

    from https://pypi.org/project/mockserver-friendly-client/

    :param folder: where to look for .json files (recursively)
    :param mock_client:
    :param resourceType:
    :param url_prefix:
    """
    file_path: str
    files: List[str] = glob(str(folder.joinpath("**/*.json")), recursive=True)
    for file_path in files:
        # load file as json
        with open(file_path, "r") as file:
            fhir_request: Dict[str, Any] = json.loads(file.read())
            # find id and resourceType
            id_: str = fhir_request["id"]
            path = f"{('/' + url_prefix) if url_prefix else ''}/4_0_0/{resourceType}/{id_}/$everything"
            mock_client.expect(
                request=mock_request(
                    method="GET",
                    path=path,
                ),
                response=mock_response(body=json.dumps(fhir_request)),
                timing=times(1),
                file_path=file_path,
            )
    return files


# noinspection PyPep8Naming
def load_mock_fhir_everything_batch_requests_from_folder(
    folder: Path,
    mock_client: MockServerFriendlyClient,
    resourceType: str,
    ids: List[str],
    url_prefix: Optional[str] = None,
) -> List[str]:
    """
    Loads all .json files from the folder and its sub-folders

    from https://pypi.org/project/mockserver-friendly-client/

    :param folder: where to look for .json files (recursively)
    :param mock_client:
    :param resourceType:
    :param url_prefix:
    :param ids: id of resources for this batch to load
    """
    file_path: str
    files: List[str] = glob(str(folder.joinpath("**/*.json")), recursive=True)
    result_bundle = {
        "resourceType": "Bundle",
        "id": "bundle-example",
        "type": "collection",
        "entry": [],
    }
    print(f"mock fhir batch request for {ids}")
    for file_path in files:
        with open(file_path, "r") as file:
            fhir_bundle: Dict[str, Any] = json.loads(file.read())
        if "entry" not in fhir_bundle:
            print(f"{file_path} has no entry property!")
            continue
        for entry in fhir_bundle["entry"]:
            id_ = entry.get("resource", {}).get("id", "")
            if id_ in ids:
                result_bundle["entry"].append(entry)  # type: ignore
    # find id and resourceType
    path = (
        f"{('/' + url_prefix) if url_prefix else ''}/4_0_0/{resourceType}/$everything"
    )
    mock_client.expect(
        request=mock_request(
            method="GET", path=path, querystring={"id": ",".join(ids)}
        ),
        response=mock_response(body=json.dumps(result_bundle)),
        timing=times(1),
        file_path=files[0] if files else None,
    )
    return files


def load_mock_elasticsearch_requests_from_folder(
    folder: Path, mock_client: MockServerFriendlyClient, index: str
) -> List[str]:
    """
    Loads all .json files from the folder and its sub-folders

    from https://pypi.org/project/mockserver-friendly-client/

    :param folder: where to look for .json files (recursively)
    :param mock_client:
    :param index:
    """
    file_path: str
    files: List[str] = glob(str(folder.joinpath("**/*.json")), recursive=True)
    for file_path in files:
        # load file as json
        with open(file_path, "r") as file:
            lines: List[str] = file.readlines()
            http_request: str = "\n".join(
                [
                    (json.dumps(json.loads(line))) if line != "\n" else ""
                    for line in lines
                ]
            )
            # noinspection PyPep8Naming
            path = f"/{index}/_bulk"
            # noinspection SpellCheckingInspection
            mock_client.expect(
                request=mock_request(
                    method="POST",
                    path=path,
                    body=text_equals(http_request),
                ),
                response=mock_response(
                    headers={"Content-Type": "application/json"},
                    body=f"""
{{
    "took": 194,
    "errors": false,
    "items": [
        {{
            "index": {{
                "_index": "{index}",
                "_type": "_doc",
                "_id": "TESQ93YBW4SQ_M9deEJw",
                "_version": 1,
                "result": "created"
            }}
        }},
        {{
            "index": {{
                "_index": "{index}",
                "_type": "_doc",
                "_id": "TUSQ93YBW4SQ_M9deEJw",
                "_version": 1,
                "result": "created"
            }}
        }}
    ]
}}""",
                ),
                timing=times(1),
                file_path=file_path,
            )
    return files


def load_mock_source_api_responses_from_folder(
    folder: Path, mock_client: MockServerFriendlyClient, url_prefix: Optional[str]
) -> List[str]:
    """
    Mock responses for all files from the folder and its sub-folders

    from https://pypi.org/project/mockserver-friendly-client/

    :param folder: where to look for files (recursively)
    :param mock_client:
    :param url_prefix:
    """
    file_path: str
    files: List[str] = sorted(glob(str(folder.joinpath("**/*")), recursive=True))
    for file_path in files:
        with open(file_path, "r") as file:
            content = file.read()
            path = f"{('/' + url_prefix) if url_prefix else ''}/{os.path.basename(file_path)}"
            mock_client.expect(
                request=mock_request(
                    method="GET",
                    path=path,
                ),
                response=mock_response(body=content),
                timing=times(1),
                file_path=file_path,
            )
    return files


def load_mock_source_api_json_responses(
    folder: Path,
    mock_client: MockServerFriendlyClient,
    url_prefix: Optional[str],
    add_file_name: Optional[bool] = False,
    url_suffix: Optional[str] = None,
) -> List[str]:
    """
    Mock responses for all files from the folder and its sub-folders

    :param folder: where to look for files (recursively)
    :param mock_client:
    :param url_prefix: http://{mock_server_url}/{url_prefix}...
    :param add_file_name: http://{mock_server_url}/{url_prefix}/{add_file_name}...
    :param url_suffix: http://{mock_server_url}/{url_prefix}/{add_file_name}/{url_suffix}?
    """
    file_path: str
    files: List[str] = sorted(glob(str(folder.joinpath("**/*.json")), recursive=True))
    for file_path in files:
        file_name = os.path.basename(file_path)
        with open(file_path, "r") as file:
            content = json.loads(file.read())

            try:
                request_parameters = content["request_parameters"]
            except ValueError:
                raise Exception(
                    "`request_parameters` key not found! It is supposed to contain parameters of the request function."
                )

            if "path" in request_parameters:
                if not request_parameters["path"].startswith("/"):
                    path = f"{('/' + url_prefix) if url_prefix else ''}"
                    path = f"{path}/{request_parameters['path']}"
                else:
                    path = request_parameters["path"]
                del request_parameters["path"]
            else:
                path = f"{('/' + url_prefix) if url_prefix else ''}"
                path = (
                    f"{path}/{os.path.splitext(file_name)[0]}"
                    if add_file_name
                    else path
                )
                if url_suffix:
                    path = f"{path}/{url_suffix}"

            code: Optional[int] = None
            try:
                request_result = content["request_result"]
                if "statusCode" in request_result:
                    code = request_result["statusCode"]

            except ValueError:
                raise Exception(
                    "`request_result` key not found. "
                    + "It is supposed to contain the expected result of the request function."
                )

            mock_client.expect(
                request=mock_request(path=path, **request_parameters),
                response=mock_response(body=json.dumps(request_result))
                if not code
                else mock_response(code=code),
                timing=times(1),
                file_path=file_path,
            )
    return files



import collections
import glob
import json
import logging
import os
import re
from logging import Logger
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, cast

import dictdiffer  # type: ignore
from requests import put, Response

from mockserver_client.exceptions.mock_server_exception import (
    MockServerException,
)
from mockserver_client.exceptions.mock_server_expectation_not_found_exception import (
    MockServerExpectationNotFoundException,
)
from mockserver_client.exceptions.mock_server_json_content_mismatch_exception import (
    MockServerJsonContentMismatchException,
)
from mockserver_client.exceptions.mock_server_request_not_found_exception import (
    MockServerRequestNotFoundException,
)
from ._time import _Time
from ._timing import _Timing
from .match_request_result import MatchRequestResult
from .mock_expectation import MockExpectation
from .mock_request import MockRequest
from .mock_request_logger import MockRequestLogger
from .mock_request_response import MockRequestResponse
from .mock_response import MockResponse
from .mockserver_verify_exception import MockServerVerifyException


class MockServerFriendlyClient(object):
    """
    Client for the MockServer
    Based on https://pypi.org/project/mockserver-friendly-client/
    """

    def __init__(
        self,
        base_url: str,
        log_all_requests_to_folder: str | Path | None = None,
        logger: Optional[Logger] = None,
    ) -> None:
        """
        Client for the MockServer
        Based on https://pypi.org/project/mockserver-friendly-client/

        :param base_url: base url to use
        """
        self.base_url: str = base_url
        self.expectations: List[MockExpectation] = []
        self.logger: Logger = logger or logging.getLogger("MockServerClient")
        if not logger:
            self.logger.setLevel(os.environ.get("LOGLEVEL") or logging.INFO)
        self.log_all_requests_to_folder: str | Path | None = log_all_requests_to_folder

    def _call(
        self, command: str, data: Any = None, query_string: Optional[str] = None
    ) -> Response:
        url = "{}/{}".format(self.base_url, command)
        if query_string:
            url += "?" + query_string
        try:
            return put(url, data=data)
        except Exception as e:
            raise Exception(f"Error calling {url}: {e}")

    def clear(self, path: str) -> None:
        """
        Clear all data related to this path


        :param path:
        """
        self.expectations = []
        self._call("clear", json.dumps({"path": path}))

    def reset(self) -> None:
        """
        Clear all data in the MockServer

        """
        self.expectations = []
        self._call("reset")

    def stub(
        self,
        *,
        request: Any,
        response: Any,
        timing: Any = None,
        time_to_live: Any = None,
    ) -> None:
        """
        Create an expectation in mock server


        :param request: mock request
        :param response: mock response
        :param timing: how many times to expect the request
        :param time_to_live:
        """
        self._call(
            "expectation",
            json.dumps(
                _non_null_options_to_dict(
                    _Option("httpRequest", request),
                    _Option("httpResponse", response),
                    _Option("times", (timing or _Timing()).for_expectation()),
                    _Option("timeToLive", time_to_live, formatter=_to_time_to_live),
                )
            ),
        )

    def expect(
        self,
        *,
        request: Dict[str, Any],
        response: Dict[str, Any],
        timing: _Timing,
        time_to_live: Any = None,
        file_path: Optional[str],
    ) -> None:
        """
        Expect this mock request and reply with the provided mock response


        :param request: mock request
        :param response: mock response
        :param timing: how many times to expect the request
        :param time_to_live:
        :param file_path: file path
        """
        self.stub(
            request=request, response=response, timing=timing, time_to_live=time_to_live
        )
        self.expectations.append(
            MockExpectation(
                request=request,
                response=response,
                timing=timing,
                index=len(self.expectations),
                file_path=file_path,
            )
        )
        MockRequestLogger.log(
            file_path=file_path,
            base_url=self.base_url,
            request=request,
            response=response,
        )

    def expect_files_as_requests(
        self,
        folder: Path,
        url_prefix: Optional[str],
        content_type: str = "application/fhir+json",
        add_file_name: bool = False,
    ) -> List[str]:
        """
        Read the files in the specified folder and create mock requests for each file


        :param folder: folder to read the files from
        :param url_prefix: url_prefix to use when mocking requests
        :param content_type: content type to use for requests
        :param add_file_name: whether to add the file name to the url
        :return: list of files read
        """
        file_path: str
        files: List[str] = sorted(
            glob.glob(str(folder.joinpath("**/*.json")), recursive=True)
        )
        for file_path in files:
            file_name = os.path.basename(file_path)
            with open(file_path, "r") as file:
                content = json.loads(file.read())

                try:
                    request_parameters = content["request_parameters"]
                except ValueError:
                    raise Exception(
                        "`request_parameters` key not found! It is supposed to contain parameters of the request function."
                    )

                path = f"{('/' + url_prefix) if url_prefix else ''}"
                path = (
                    f"{path}/{os.path.splitext(file_name)[0]}"
                    if add_file_name
                    else path
                )

                try:
                    request_result = content["request_result"]
                except ValueError:
                    raise Exception(
                        "`request_result` key not found. It is supposed to contain the expected result of the requst function."
                    )
                body = (
                    json.dumps(request_result)
                    if content_type == "application/fhir+json"
                    else request_result
                )
                self.expect(
                    request=mock_request(path=path, **request_parameters),
                    response=mock_response(body=body),
                    timing=times(1),
                    file_path=file_path,
                )
                self.logger.info(f"Mocking {self.base_url}{path}: {request_parameters}")
        return files

    def expect_files_as_json_requests(
        self,
        folder: Path,
        path: str,
        json_response_body: Dict[str, Any],
        add_file_name: bool = False,
    ) -> List[str]:
        """
        Read the files in the specified folder and create mock requests for each file


        :param folder: folder to read the files from
        :param path: url_prefix to use when mocking requests
        :param json_response_body: mock response body to return for each mock request
        :param add_file_name: whether to add the file name to the url
        :return: list of files read
        """
        file_path: str
        files: List[str] = sorted(
            glob.glob(str(folder.joinpath("**/*.json")), recursive=True)
        )
        for file_path in files:
            file_name = os.path.basename(file_path)
            with open(file_path, "r") as file:
                content: Dict[str, Any] = json.loads(file.read())
                path = (
                    f"{path}/{os.path.splitext(file_name)[0]}"
                    if add_file_name
                    else path
                )
                self.expect(
                    request=mock_request(
                        path=path, body=json_equals([content]), method="POST"
                    ),
                    response=mock_response(body=json.dumps(json_response_body)),
                    timing=times(1),
                    file_path=file_path,
                )
                self.logger.info(f"Mocking {self.base_url}{path}")
        return files

    def expect_default(
        self,
    ) -> None:
        """
        Fallback handler for all requests


        """
        response: Dict[str, Any] = mock_response()
        timing: _Timing = times_any()
        self.stub(request={}, response=response, timing=timing, time_to_live=None)
        self.expectations.append(
            MockExpectation(
                {}, {}, timing, index=len(self.expectations), file_path="{catch all}"
            )
        )

    def match_to_recorded_requests(
        self,
        recorded_requests: List[MockRequest],
    ) -> MatchRequestResult:
        """
        Matches recorded requests with expected requests
        There are 4 cases possible:
        1. There was an expectation without a corresponding request -> fail
        2. There was a request without a corresponding expectation -> save request as expectation
        3. There was a matching request and expectation but the content did not match -> error and show diff
        4. There was a matching request and expectation and the content matched -> nothing to do


        :param recorded_requests: list of requests actually made to the mock server
        :return: list of match exceptions
        """
        exceptions: List[MockServerException] = []
        unmatched_expectation_requests: List[MockRequest] = []
        unmatched_requests: List[MockRequest] = [r for r in recorded_requests]
        expected_request: MockRequest
        self.logger.debug("-------- EXPECTATIONS --------")
        expectation: MockExpectation
        for expectation in self.expectations:
            self.logger.debug(expectation)
        self.logger.debug("-------- END EXPECTATIONS --------")
        self.logger.debug("-------- REQUESTS --------")
        recorded_request: MockRequest
        for recorded_request in recorded_requests:
            self.logger.debug(recorded_request)
        self.logger.debug("-------- END REQUESTS --------")

        # get ids of all recorded requests
        recorded_request_ids: List[str] = []
        for recorded_request in recorded_requests:
            json1: Optional[List[Dict[str, Any]]] = recorded_request.json_list
            if json1:
                # get ids from body and match
                # see if the property is string
                # noinspection PyTypeChecker
                json1_id_list: List[str] = [j["id"] for j in json1 if "id" in j]
                for j in json1_id_list:
                    recorded_request_ids.append(j)

        matched_requests: List[MockRequest] = []
        self.logger.info(f"========= START MATCHING EXPECTATIONS  ================")
        # now try to match requests to expectations
        for expectation in self.expectations:
            expected_request = expectation.request
            self.logger.info(
                f"------- Expectation {expected_request.index}/{len(self.expectations) - 1} -------"
            )
            self.logger.info(f"{expected_request}")
            matching_request: Optional[MockRequest] = None
            recorded_requests_not_matched_yet: List[MockRequest] = [
                r for r in recorded_requests if not r in matched_requests
            ]
            try:
                matching_request = self.find_matches_on_request_and_body(
                    expected_request=expected_request,
                    recorded_requests=recorded_requests_not_matched_yet,
                    unmatched_requests=unmatched_requests,
                )
                if matching_request:
                    self.logger.info(f"MATCHED (exact) to {matching_request}")
                else:
                    matching_request = self.find_matches_on_request_url_only(
                        expected_request=expected_request,
                        recorded_requests=recorded_requests_not_matched_yet,
                        unmatched_requests=unmatched_requests,
                    )
                    if matching_request:
                        matched_requests.append(matching_request)
                        self.logger.info(f"MATCHED (url only) to {matching_request}")
                    else:
                        self.logger.info(f"NO {matching_request}")
            except MockServerJsonContentMismatchException as e:
                exceptions.append(e)
            if not matching_request and expected_request.method:
                unmatched_expectation_requests.append(expected_request)
                self.logger.info("---- EXPECTATION NOT MATCHED ----")
                self.logger.info(f"{expected_request}")
                self.logger.info("IDs sent in requests")
                self.logger.info(f'{",".join(recorded_request_ids)}')
                self.logger.info("---- END EXPECTATION NOT MATCHED ----")
        self.logger.info(f"========= END MATCHING EXPECTATIONS ================")

        # now fail for every expectation in unmatched_expectation_requests
        for unmatched_expectation in unmatched_expectation_requests:
            exceptions.append(
                MockServerExpectationNotFoundException(
                    method=unmatched_expectation.method,
                    url=unmatched_expectation.path,
                    json_list=unmatched_expectation.json_list,
                    querystring_params=unmatched_expectation.querystring_params,
                    expectation=unmatched_expectation,
                )
            )
        # and for every request in unmatched_requests
        for unmatched_request in unmatched_requests:
            exceptions.append(
                MockServerRequestNotFoundException(
                    method=unmatched_request.method,
                    url=unmatched_request.path,
                    querystring_params=unmatched_request.querystring_params,
                    json_list=unmatched_request.json_list,
                    request=unmatched_request,
                )
            )
        return MatchRequestResult(
            exceptions=exceptions, found_expectations=matched_requests
        )

    def find_matches_on_request_url_only(
        self,
        *,
        expected_request: MockRequest,
        recorded_requests: List[MockRequest],
        unmatched_requests: List[MockRequest],
    ) -> Optional[MockRequest]:
        """
        Finds matches on url only and then compares the bodies.  Returns if match was found.
        Throws a JsonContentMismatchException if a url match was found but no body match was found


        :param expected_request: request that was expected
        :param recorded_requests: list of all requests made
        :param unmatched_requests: list of requests that have not been matched to an expectation
        :return: whether a matching expectation was found
        """
        matched_request: Optional[MockRequest] = None
        recorded_request: MockRequest
        for recorded_request in recorded_requests:
            matched_request = self.does_request_match_on_url_only(
                expected_request=expected_request,
                recorded_request=recorded_request,
                unmatched_requests=unmatched_requests,
            )
            if matched_request:
                return matched_request
        return matched_request

    def does_request_match_on_url_only(
        self,
        *,
        expected_request: MockRequest,
        recorded_request: MockRequest,
        unmatched_requests: List[MockRequest],
    ) -> Optional[MockRequest]:
        """
        Checks if the two requests match on url only


        """
        request_matched = expected_request.method and self.does_request_match(
            request1=expected_request, request2=recorded_request, check_body=False
        )
        request_id_matched = (
            expected_request.json_list is not None
            and recorded_request.json_list is not None
            and self.does_id_in_request_match(
                request1=expected_request, request2=recorded_request
            )
        )
        if request_matched or request_id_matched:
            # find all requests that match on url since there can be multiple
            # and then check if the bodies match
            # matching_expectations = [
            #     m
            #     for m in self.expectations
            #     if "method" in m.request
            #     and self.does_request_match(
            #         request1=m.request,
            #         request2=recorded_request,
            #         check_body=False,
            #     )
            # ]
            matched_request: MockRequest = recorded_request
            # remove request from unmatched_requests
            unmatched_request_list = [
                r
                for r in unmatched_requests
                if self.does_request_match(
                    request1=r, request2=recorded_request, check_body=True
                )
            ]
            if expected_request.json_list:
                expected_body_json: Optional[
                    List[Dict[str, Any]]
                ] = expected_request.json_list
                actual_body_json: Optional[
                    List[Dict[str, Any]]
                ] = recorded_request.json_list
                assert len(unmatched_request_list) < 2, (
                    f"Found {len(unmatched_request_list)}"
                    f" unmatched requests for {recorded_request}"
                )
                if len(unmatched_request_list) > 0:
                    unmatched_requests.remove(unmatched_request_list[0])
                self.compare_request_bodies_json(actual_body_json, expected_body_json)
            elif expected_request.body_list:
                if len(unmatched_request_list) > 0:
                    unmatched_requests.remove(unmatched_request_list[0])
                self.compare_request_bodies(
                    recorded_request.body_list, expected_request.body_list
                )
            return matched_request
        return None

    def find_matches_on_request_and_body(
        self,
        *,
        expected_request: MockRequest,
        recorded_requests: List[MockRequest],
        unmatched_requests: List[MockRequest],
    ) -> Optional[MockRequest]:
        """
        Matches on both request and body and returns whether it was able to find a match


        :param expected_request: request that was expected
        :param recorded_requests: list of all requests made
        :param unmatched_requests: list of requests that have not been matched to an expectation
        :return: whether a matching expectation was found
        """
        # first try to find all exact matches on both request url and body
        matching_request: Optional[MockRequest] = None
        for recorded_request in recorded_requests:
            matching_request = self.does_request_match_on_url_and_body(
                expected_request=expected_request,
                recorded_request=recorded_request,
                unmatched_requests=unmatched_requests,
            )
            if matching_request:
                return matching_request
            # now try to find matches on just url
        return matching_request

    def does_request_match_on_url_and_body(
        self,
        *,
        expected_request: MockRequest,
        recorded_request: MockRequest,
        unmatched_requests: List[MockRequest],
    ) -> Optional[MockRequest]:
        """
        Returns the request if it matches and removes it from the unmatched_requests list


        """
        # first try to match on both request url AND body
        # If match is found then remove this request from list of unmatched requests
        if expected_request.method and self.does_request_match(
            request1=expected_request,
            request2=recorded_request,
            check_body=True,
        ):
            matching_request = recorded_request
            # remove request from unmatched_requests
            unmatched_request_list = [
                r
                for r in unmatched_requests
                if self.does_request_match(
                    request1=r, request2=recorded_request, check_body=True
                )
            ]
            assert (
                len(unmatched_request_list) >= 0
            ), f"{','.join([str(c) for c in unmatched_request_list])}"
            if len(unmatched_request_list) > 0:
                unmatched_requests.remove(unmatched_request_list[0])
            return matching_request
        return None

    @staticmethod
    def does_request_match(
        request1: MockRequest, request2: MockRequest, check_body: bool
    ) -> bool:
        """
        Does request1 match request2

        :param request1: request 1
        :param request2: request 2
        :param check_body: whether to match the body or not
        :return: whether the two requests match
        """
        if request1.method != request2.method:
            return False
        if request1.path != request2.path:
            return False
        request1_query_string: Optional[
            Dict[str, Any]
        ] = MockServerFriendlyClient.normalize_querystring_params(
            request1.querystring_params
        )
        request2_query_string: Optional[
            Dict[str, Any]
        ] = MockServerFriendlyClient.normalize_querystring_params(
            request2.querystring_params
        )
        if request1_query_string != request2_query_string:
            return False
        if (
            request1.json_list is not None
            and request2.json_list is not None
            and not MockServerFriendlyClient.does_id_in_request_match(
                request1=request1, request2=request2
            )
        ):
            return False
        if check_body and not MockServerFriendlyClient.does_request_body_match(
            request1=request1, request2=request2
        ):
            return False
        return True

    @staticmethod
    def does_request_body_match(request1: MockRequest, request2: MockRequest) -> bool:
        """
        Does the body of the two specified requests match

        :param request1: request 1
        :param request2: request 2
        :return: whether the body of the two specified requests match
        :rtype:
        """
        if not request1.body_list and not request2.body_list:
            return True
        if request1.body_list and not request2.body_list:
            return False
        if request2.body_list and not request1.body_list:
            return False
        if request1.json_list and request2.json_list:
            # now compare non bundle resources
            comparison_results = list(
                dictdiffer.diff(request1.json_list, request2.json_list)
            )
            return True if len(comparison_results) == 0 else False
        return True if request1.body_list == request2.body_list else False

    @staticmethod
    def does_id_in_request_match(request1: MockRequest, request2: MockRequest) -> bool:
        """
        Whether the id in the two specified requests match.


        :param request1: request 1
        :param request2: request 2
        :return: Whether the id in the two specified requests match.
        """
        json1_list: Optional[List[Dict[str, Any]]] = request1.json_list
        json2_list: Optional[List[Dict[str, Any]]] = request2.json_list

        if json1_list and json2_list:
            # handle bundles
            if len(json1_list) == 1 and len(json2_list) == 1:
                request1_first_json: Dict[str, Any] = json1_list[0]
                request2_first_json: Dict[str, Any] = json2_list[0]
                if (
                    request1_first_json.get("resourceType")
                    == request2_first_json.get("resourceType")
                    == "Bundle"
                ):
                    request1_entries: List[
                        Dict[str, Any]
                    ] | None = request1_first_json.get("entry")
                    request2_entries: List[
                        Dict[str, Any]
                    ] | None = request2_first_json.get("entry")
                    if (
                        request1_entries
                        and len(request1_entries) > 0
                        and request2_entries
                        and len(request2_entries) > 0
                    ):
                        request1_first_resource: Dict[
                            str, Any
                        ] | None = request1_entries[0].get("resource")
                        request2_first_resource: Dict[
                            str, Any
                        ] | None = request2_entries[0].get("resource")
                        if request1_first_resource and request2_first_resource:
                            request1_first_resource_id: Optional[
                                str
                            ] = request1_first_resource.get("id")
                            request2_first_resource_id: Optional[
                                str
                            ] = request2_first_resource.get("id")
                            if (
                                request1_first_resource_id != request2_first_resource_id
                                or request1_first_resource.get("resourceType")
                                != request2_first_resource.get("resourceType")
                            ):
                                return False
            # get ids from body and match
            # see if the property is string
            json1_id_list: List[str] = [j["id"] for j in json1_list if "id" in j]
            json1_resource_type_list: List[str] = [
                j["resourceType"] for j in json1_list if "resourceType" in j
            ]
            json2_id_list: List[str] = [j["id"] for j in json2_list if "id" in j]
            json2_resource_type_list: List[str] = [
                j["resourceType"] for j in json2_list if "resourceType" in j
            ]

            return (
                True
                if json1_id_list == json2_id_list
                and json1_resource_type_list == json2_resource_type_list
                else False
            )
        elif json1_list is None and json2_list is None:
            return True
        else:
            return False

    @staticmethod
    def compare_request_bodies(
        actual_body_list: Optional[List[Dict[str, Any]]],
        expected_body_list: Optional[List[Dict[str, Any]]],
    ) -> None:
        """
        Compares the bodies of the two requests and raises an exception with detailed diff if they don't match


        :param actual_body_list: body of actual request
        :param expected_body_list: body of expected request
        """
        differences = list(dictdiffer.diff(expected_body_list, actual_body_list))
        if len(differences) > 0:
            raise MockServerJsonContentMismatchException(
                actual_json=actual_body_list,
                expected_json=expected_body_list,
                differences=differences,
                expected_file_path=Path(),
            )

    @staticmethod
    def compare_request_bodies_json(
        actual_json: Optional[List[Dict[str, Any]]],
        expected_json: Optional[List[Dict[str, Any]]],
    ) -> None:
        """
        Compares the JSON bodies of the two requests and raises an exception with detailed diff if they don't match

        :param actual_json: json of actual request
        :param expected_json: json of expected request
        """
        differences = list(dictdiffer.diff(expected_json, actual_json))
        if len(differences) > 0:
            raise MockServerJsonContentMismatchException(
                actual_json=actual_json,
                expected_json=expected_json,
                differences=differences,
                expected_file_path=Path(),
            )

    def verify_expectations(
        self, test_name: Optional[str] = None, files: Optional[List[str]] = None
    ) -> None:
        """
        Verify that the requests made match the expectations.  Raises exceptions if there are mismatches


        :param test_name: Name of test
        :param files: files to create expectations
        """
        recorded_requests: List[MockRequest] = self.retrieve_requests()
        recorded_request_responses: List[
            MockRequestResponse
        ] = self.retrieve_request_responses()
        self.logger.debug(f"Count of retrieved requests: {len(recorded_requests)}")
        if self.log_all_requests_to_folder:
            self.write_all_requests_to_folder(
                request_responses=recorded_request_responses
            )
        self.logger.debug("-------- All Retrieved Requests -----")
        for recorded_request in recorded_requests:
            self.logger.debug(f"{recorded_request}")
        self.logger.debug("-------- End All Retrieved Requests -----")
        # now filter to the requests for this test only
        if test_name is not None:
            recorded_requests = [
                r for r in recorded_requests if r.path and test_name in r.path
            ]
        self.logger.debug(
            f"Count of recorded requests for test: {len(recorded_requests)}"
        )
        match_result: MatchRequestResult = self.match_to_recorded_requests(
            recorded_requests=recorded_requests
        )
        exceptions: List[MockServerException] = match_result.exceptions
        found_expectations: List[MockRequest] = match_result.found_expectations

        if len(exceptions) > 0:
            self.logger.info("-------- Matched Retrieved Requests -----")
            for found_expectation in found_expectations:
                self.logger.info(f"{found_expectation}")
            self.logger.info("-------- End Matched Retrieved Requests -----")
            raise MockServerVerifyException(
                exceptions=exceptions,
                files=files,
                found_expectations=found_expectations,
            )

    def retrieve_requests(self) -> List[MockRequest]:
        """
        Retrieve requests made to mock server


        :return: list of requests made to mock server
        """
        result = self._call("retrieve")
        # https://app.swaggerhub.com/apis/jamesdbloom/mock-server-openapi/5.11.x#/control/put_retrieve
        raw_requests: List[Dict[str, Any]] = cast(
            List[Dict[str, Any]], json.loads(result.text)
        )
        return [
            MockRequest(request=r, index=index, file_path=None)
            for index, r in enumerate(raw_requests)
        ]

    def retrieve_request_responses(self) -> List[MockRequestResponse]:
        """
        Retrieve requests made to mock server


        :return: list of requests made to mock server
        """
        result = self._call("retrieve", query_string="type=request_responses")
        # https://app.swaggerhub.com/apis/jamesdbloom/mock-server-openapi/5.11.x#/control/put_retrieve
        raw_requests: List[Dict[str, Any]] = cast(
            List[Dict[str, Any]], json.loads(result.text)
        )
        return [
            MockRequestResponse(
                request=r.get("httpRequest"),
                response=r.get("httpResponse"),
                index=index,
            )
            for index, r in enumerate(raw_requests)
        ]

    @staticmethod
    def safe_string_for_file_path(s: str) -> str:
        # Replace spaces with underscores
        s = s.replace(" ", "_")

        # replace / with +
        s = s.replace("/", "+")

        # Convert to lowercase
        s = s.lower()

        # Remove any remaining non-alphanumeric characters
        s = re.sub(r"[^a-z0-9_+]", "", s)

        # Limit length if needed
        max_length = 255  # Most file systems have a 255-character limit for file names
        if len(s) > max_length:
            s = s[:max_length]

        return s

    def write_all_requests_to_folder(
        self, request_responses: List[MockRequestResponse]
    ) -> None:
        assert self.log_all_requests_to_folder
        # write all requests to file
        recorded_request_response: MockRequestResponse
        for index, recorded_request_response in enumerate(request_responses):
            request: MockRequest | None = recorded_request_response.request
            if not request:
                continue
            json_dict: Dict[str, Any] = {
                "request_parameters": {
                    "method": f"{request.method}",
                    "path": request.path,
                }
            }
            if request.querystring_params:
                json_dict["request_parameters"][
                    "querystring"
                ] = request.querystring_params
            if request.json_list:
                json_dict["request_body"] = request.json_list
            response: MockResponse | None = recorded_request_response.response
            if response:
                if response.json_body:
                    json_dict["request_result"] = response.json_body
                elif response.status_code:
                    json_dict["request_result"] = {"status_code": response.status_code}

            json_content = json.dumps(json_dict, indent=4)
            # path_parts: List[str] = recorded_request_response.path.split("/")
            file_name: str = (
                f"{index}-{self.safe_string_for_file_path(request.path)}.json"
                if request.path
                else f"{index}.json"
            )
            path = Path(self.log_all_requests_to_folder)
            # for path_part in path_parts:
            #     path = path.joinpath(path_part)
            path = path.joinpath(file_name)
            with open(path, "w") as file:
                file.write(json_content)

    @staticmethod
    def normalize_querystring_params(
        querystring_params: Optional[Union[List[Dict[str, Any]], Dict[str, Any]]] = None
    ) -> Optional[Dict[str, Any]]:
        """
        ensure a dictionary of querystring params is formatted so that the param name is the dictionary key.
        querystring dictionaries from requests sometimes look like this. don't want that.
        "queryStringParameters": [
            {
                "name": "contained",
                "values": [
                    "true"
                ]
            },
            {
                "name": "id",
                "values": [
                    "1023011178"
                ]
            }
        ],
        """
        if querystring_params is None:
            return None
        if type(querystring_params) is dict:
            return querystring_params

        normalized_params: Dict[str, Any] = {}
        for param_dict in querystring_params:
            params: Dict[str, Any] = param_dict  # type: ignore
            normalized_params[params["name"]] = params["values"]
        return normalized_params


def mock_request(
    method: Optional[str] = None,
    path: Optional[str] = None,
    querystring: Optional[Dict[str, Any]] = None,
    body: Optional[Union[str, Dict[str, Any]]] = None,
    headers: Optional[Dict[str, Any]] = None,
    cookies: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Mocks a request


    :param method: method of request
    :param path: path of request (No query strings in path.  Use the querystring parameter)
    :param querystring: dict of query strings
    :param body: body to expect in the request
    :param headers: headers to expect in the request
    :param cookies: cookies to expect in the request
    :return: mock request
    """
    assert (
        not path or "?" not in path
    ), "Do not specify query string in the path.  Use the querystring parameter"

    return _non_null_options_to_dict(
        _Option("method", method),
        _Option("path", path),
        _Option("queryStringParameters", querystring, formatter=_to_named_values_list),
        _Option("body", body),
        _Option("headers", headers, formatter=_to_named_values_list),
        _Option("cookies", cookies),
    )


def mock_response(
    code: Optional[int] = None,
    body: Optional[Union[str, Dict[str, Any]]] = None,
    headers: Optional[Dict[str, Any]] = None,
    cookies: Optional[str] = None,
    delay: Optional[str] = None,
    reason: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Specifies the mock response for a mock request


    :param code: code to return in mock response
    :param body: body to return in mock response
    :param headers: headers to return in mock response
    :param cookies: cookies to return in mock response
    :param delay: delay to use before returning response
    :param reason: reason phrase to return in mock response
    :return: mock response
    """
    return _non_null_options_to_dict(
        _Option("statusCode", code),
        _Option("reasonPhrase", reason),
        _Option("body", body),
        _Option("headers", headers, formatter=_to_named_values_list),
        _Option("delay", delay, formatter=_to_delay),
        _Option("cookies", cookies),
    )


def times(count: int) -> _Timing:
    """
    How many times to expect the request


    :param count: count
    :return: Timing object
    """
    return _Timing(count)


def times_once() -> _Timing:
    """
    Expect the request a single time


    :return: Timing object
    """
    return _Timing(1)


def times_any() -> _Timing:
    """
    Expect the request unlimited number of times


    :return: Timing object
    """
    return _Timing()


def form(form1: Any) -> Dict[str, Any]:
    # NOTE(lindycoder): Support for mockservers version before https://github.com/jamesdbloom/mockserver/issues/371
    return collections.OrderedDict(
        (("type", "PARAMETERS"), ("parameters", _to_named_values_list(form1)))
    )


def json_equals(payload: Union[Dict[str, Any], List[Dict[str, Any]]]) -> Dict[str, Any]:
    """
    Expects that the request payload is equal to the given payload.

    :param payload: json to compare to
    :return:
    """
    return collections.OrderedDict(
        (("type", "JSON"), ("json", json.dumps(payload)), ("matchType", "STRICT"))
    )


def text_equals(payload: str) -> Dict[str, Any]:
    """
    Expects that the request payload is equal to the given payload.

    :param payload: text to compare to
    :return:
    """
    return collections.OrderedDict(
        (
            ("type", "STRING"),
            ("string", payload),
            ("contentType", "text/plain; charset=utf-8"),
        )
    )


def json_contains(payload: Dict[str, Any]) -> Dict[str, Any]:
    """
    Expects the request payload to match all given fields. The request may have more fields.


    :param payload: returned json must include this
    :return:
    """
    return collections.OrderedDict(
        (
            ("type", "JSON"),
            ("json", json.dumps(payload)),
            ("matchType", "ONLY_MATCHING_FIELDS"),
        )
    )


def json_response(
    body: Any = None, headers: Any = None, **kwargs: Any
) -> Dict[str, Any]:
    """
    Expect this json response


    :param body:
    :param headers:
    :param kwargs:
    :return:
    """
    headers = headers or {}
    headers["Content-Type"] = "application/json"
    return mock_response(body=json.dumps(body), headers=headers, **kwargs)


class _Option:
    def __init__(self, field: Any, value: Any, formatter: Any = None) -> None:
        self.field = field
        self.value = value
        self.formatter = formatter or (lambda e: e)


def seconds(value: int) -> _Time:
    return _Time("SECONDS", value)


def milliseconds(value: int) -> _Time:
    return _Time("MILLISECONDS", value)


def microseconds(value: int) -> _Time:
    return _Time("MICROSECONDS", value)


def nanoseconds(value: int) -> _Time:
    return _Time("NANOSECONDS", value)


def minutes(value: int) -> _Time:
    return _Time("MINUTES", value)


def hours(value: int) -> _Time:
    return _Time("HOURS", value)


def days(value: int) -> _Time:
    return _Time("DAYS", value)


def _non_null_options_to_dict(*options: Any) -> Dict[str, Any]:
    return {o.field: o.formatter(o.value) for o in options if o.value is not None}


def _to_named_values_list(dictionary: Dict[str, Any]) -> List[Dict[str, Any]]:
    return [
        {"name": key, "values": [value] if not isinstance(value, list) else value}
        for key, value in dictionary.items()
    ]


def _to_time(value: Union[_Time, int]) -> _Time:
    if not isinstance(value, _Time):
        value = seconds(value)
    return value


def _to_delay(delay: _Time) -> Dict[str, Any]:
    delay = _to_time(delay)
    return {"timeUnit": delay.unit, "value": delay.value}


def _to_time_to_live(time: Union[_Time, int]) -> Dict[str, Any]:
    time = _to_time(time)
    return {"timeToLive": time.value, "timeUnit": time.unit, "unlimited": False}



from typing import Any, Dict, Optional

from mockserver_client._timing import _Timing
from mockserver_client.mock_request import MockRequest


class MockExpectation:
    def __init__(
        self,
        request: Dict[str, Any],
        response: Dict[str, Any],
        timing: _Timing,
        index: int,
        file_path: Optional[str],
    ) -> None:
        """
        Class for Expectation

        :param request: request
        :param response: response
        :param timing: timing
        """
        self.request: MockRequest = MockRequest(
            request=request, index=index, file_path=file_path
        )
        self.response: Dict[str, Any] = response
        self.timing: _Timing = timing

    def __str__(self) -> str:
        return str(self.request)
